# HTTP server for Atom feed of TLS certificate expirations
#
# SPDX-FileCopyrightText: 2025 Nguyễn Gia Phong
# SPDX-License-Identifier: AGPL-3.0-or-later

from argparse import ArgumentParser
from asyncio import run, start_server
from base64 import urlsafe_b64decode as from_base64
from datetime import datetime, timezone
from functools import lru_cache, partial
from http import HTTPStatus
from operator import call
from pathlib import Path
from string import digits
from typing import assert_never
from urllib.parse import parse_qs, urljoin, urlsplit
from xml.etree.ElementTree import (Element as xml_element,
                                   SubElement as xml_subelement,
                                   indent, tostring as string_from_xml)
from sys import argv

from . import (__version__, GNUHelpFormatter, NetLoc,
               format_epilog, format_version)
from .check import base64_from_str

__all__ = ['main']


@lru_cache
def read_text(file, time):
    """Read the given text file if it has been updated."""
    return file.read_text()


def datetime_from_str(string, unavailable_ok=False):
    """Parse datetime from string in ISO 8601 format."""
    if string == 'N/A' and unavailable_ok:
        return None
    return datetime.fromisoformat(string)


def str_from_base64(string64):
    """Decode string in base64 format."""
    return from_base64(string64.encode()).decode()


def parse_summary(line):
    """Parse TLS certificate into a summary tuple."""
    return tuple(map(call,
                     (partial(datetime_from_str, unavailable_ok=True),
                      datetime_from_str, str, int, int, str_from_base64),
                     line.rstrip('\r\n').split(' ', maxsplit=5)))


def path(hostname, port, number, string):
    """Return the relative URL for the given certificate's details."""
    return f'{hostname}/{port}/{base64_from_str(string)}/{number}'


def supported_http_version(version):
    """Check if given HTTP version complies with section 2.5 of RFC 9110."""
    match len(version):
        case 1:
            return version in digits
        case 3:
            major, period, minor = version
            return major in digits and period == '.' and minor in digits
        case _:
            return False


def write_status(writer, http_version, status):
    """Write the given HTTP status line."""
    status = f'HTTP/{http_version} {status.value} {status.phrase}\r\n'
    writer.write(status.encode())


def write_content_type(writer, content_type):
    """Write the given HTTP content type."""
    writer.write(f'Content-Type: {content_type}\r\n'.encode())


def describe_status(writer, status, http_version='1.1'):
    """Write a HTTP/1.1 response including status description."""
    write_status(writer, http_version, status)
    content = f'{status.description}\n'.encode()
    write_content_type(writer, 'text/plain')
    writer.write(f'Content-Length: {len(content)}\r\n\r\n'.encode())
    writer.write(content)


def body(not_before, not_after, hostname, port, number, string):
    """Describe the given certificate in XHTML."""
    if not_before is None:
        return (('h1', 'TLS certificate problem'),
                ('dl',
                 ('dt', 'Domain'), ('dd', hostname),
                 ('dt', 'Port'), ('dd', port),
                 ('dt', 'Time'), ('dd', not_after),
                 ('dt', 'Error'), ('dd', string)))
    return (('h1', 'TLS certificate information'),
            ('dl',
             ('dt', 'Domain'), ('dd', hostname),
             ('dt', 'Port'), ('dd', port),
             ('dt', 'Issuer'), ('dd', string),
             ('dt', 'Serial number'), ('dd', number),
             ('dt', 'Valid from'), ('dd', not_before),
             ('dt', 'Valid until'), ('dd', not_after)))


def entry(base_url, cert):
    """Construct Atom entry for the given TLS certificate."""
    not_before, not_after, hostname, port, number, string = cert
    url = urljoin(base_url, path(hostname, port, number, string))
    title = (f'TLS cert for {hostname} cannot be retrieved'
             if not_before is None
             else f'TLS cert for {hostname} will expire at {not_after}')
    author = 'Scadere' if not_before is None else string
    updated = not_after if not_before is None else not_before
    return ('entry',
            ('author', ('name', author)),
            ('content', {'type': 'xhtml'},
             ('div', {'xmlns': 'http://www.w3.org/1999/xhtml'}, *body(*cert))),
            ('id', url),
            ('link', {'rel': 'alternate',
                      'type': 'application/xhtml+xml',
                      'href': url}),
            ('title', title),
            ('updated', updated))


def split_domain(domain):
    """Split domain and order by ascending level."""
    return tuple(domain.split('.')[::-1])


def is_subdomain(subject, objects):
    """Check if subject is a subdomain of any object."""
    if not objects:
        return True
    sbj_parts = split_domain(subject)
    return any(sbj_parts[:len(obj_parts)] == obj_parts
               for obj_parts in map(split_domain, objects))


def feed(base_url, name, mtime, certificates, domains):
    """Construct an Atom feed based on the given information."""
    return ('feed', {'xmlns': 'http://www.w3.org/2005/Atom'},
            ('id', base_url),
            ('link', {'rel': 'self', 'href': base_url}),
            ('title', name),
            ('updated', mtime),
            ('generator',
             {'uri': 'https://trong.loang.net/scadere/about',
              'version': __version__},
             'Scadere'),
            *(entry(base_url, cert) for cert in certificates
              if is_subdomain(cert[2], domains)))


def page(certificate):
    """Construct an XHTML page for the given TLS certificate."""
    hostname, port = certificate[2:4]
    return ('html', {'xmlns': 'http://www.w3.org/1999/xhtml',
                     'lang': 'en'},
            ('head',
             ('meta', {'name': 'color-scheme',
                       'content': 'dark light'}),
             ('meta', {'name': 'viewport',
                       'content': ('width=device-width,'
                                   'initial-scale=1.0')}),
             ('link', {'rel': 'icon', 'href': 'data:,'}),
             ('title', f'TLS certificate - {hostname}:{port}')),
            ('body', *body(*certificate)))


def xml(tree, parent=None):
    """Construct XML element from the given tree."""
    tag, attrs, children = ((tree[0], tree[1], tree[2:])
                            if isinstance(tree[1], dict)
                            else (tree[0], {}, tree[1:]))
    if parent is None:
        elem = xml_element(tag, attrs)
    else:
        elem = xml_subelement(parent, tag, attrs)
    for child in children:
        match child:
            case tuple():
                xml(child, elem)
            case str():
                elem.text = child
            case int():
                elem.text = str(child)
            case datetime():
                elem.text = child.isoformat()
            case _:
                assert_never(child)
    if parent is None:
        indent(elem)
    return elem


@lru_cache
def unparsed_feed(*args):
    """Cache Atom feed."""
    return string_from_xml(xml(feed(*args)), 'unicode',
                           xml_declaration=True, default_namespace=None)


@lru_cache
def unparsed_page(*args):
    """Cache XHTML page."""
    return string_from_xml(xml(page(*args)), 'unicode',
                           xml_declaration=True, default_namespace=None)


def write_xml(writer, http_version, application, func, *args):
    """Write given document as XML."""
    try:
        content = func(*args).encode()
    except Exception:  # pragma: no cover
        describe_status(writer, HTTPStatus.INTERNAL_SERVER_ERROR, http_version)
        raise
    else:
        write_status(writer, http_version, HTTPStatus.OK)
        write_content_type(writer, f'application/{application}+xml')
        writer.write(f'Content-Length: {len(content)}\r\n\r\n'.encode())
        writer.write(content)


async def handle(certs, base_url, reader, writer, title=None):
    """Handle HTTP request."""
    try:
        try:
            request = await reader.readuntil(b'\r\n')
        except Exception:
            describe_status(writer, HTTPStatus.BAD_REQUEST)
            return

        if not request.startswith(b'GET '):
            describe_status(writer, HTTPStatus.METHOD_NOT_ALLOWED)
            return

        try:
            # Raise ValueError on the lack of b'HTTP/'
            url, version = request.removeprefix(b'GET ').rsplit(b' HTTP/', 1)
            http_version = version.strip().decode()
            if not supported_http_version(http_version):
                raise ValueError
        except ValueError:
            describe_status(writer, HTTPStatus.HTTP_VERSION_NOT_SUPPORTED)
            return

        try:
            mtime = datetime.fromtimestamp(certs.stat().st_mtime, timezone.utc)
            summaries = tuple(map(parse_summary,
                                  read_text(certs, mtime).splitlines()))
            paths = tuple(urlsplit(urljoin(base_url, path(*s[-4:]))).path
                          for s in summaries)
            lookup = dict(map(tuple, zip(paths, summaries)))
            assert len(lookup) == len(summaries)
            url_parts = urlsplit(urljoin(base_url, url.strip().decode()))
            domains = tuple(parse_qs(url_parts.query).get('domain', []))
            if title is None:
                title = certs.name
        except Exception:  # pragma: no cover
            describe_status(writer, HTTPStatus.INTERNAL_SERVER_ERROR,
                            http_version)
            raise

        if url_parts.path == urlsplit(base_url).path:  # Atom feed
            write_xml(writer, http_version, 'atom', unparsed_feed,
                      base_url, title, mtime, summaries, domains)
        elif url_parts.path in lookup:  # accessible Atom entry's link/ID
            write_xml(writer, http_version, 'xhtml', unparsed_page,
                      lookup.get(url_parts.path))
        else:
            describe_status(writer, HTTPStatus.NOT_FOUND, http_version)
    finally:
        assert writer.can_write_eof()
        writer.write_eof()
        writer.close()
        await writer.wait_closed()


async def listen(certs, title, base_url, host, port):  # pragma: no cover
    """Serve HTTP server for TLS certificate expirations' Atom feed."""
    server = await start_server(partial(handle, certs, base_url, title=title),
                                host, port)
    async with server:
        print('Serving on', end=' ')
        print(*(socket.getsockname() for socket in server.sockets), sep=', ')
        await server.serve_forever()


def with_trailing_slash(base_url):
    """Return the base URL with a trailing slash."""
    return base_url if base_url.endswith('/') else f'{base_url}/'


def main(prog=argv[0], arguments=argv[1:]):
    """Launch server."""
    desc = ('Serve at URL Atom feeds for TLS certificate renewal reminder.'
            '  It is possible for clients to filter domains'
            ' using one or more "domain" URL queries.\n\n'
            'The certificate information is read from the file at PATH,'
            ' which is generated by scadere-check(1).\n\n'
            'The server listens for TCP connections coming to HOST:PORT,'
            ' where HOST defaults to localhost'
            ' and PORT is selected randomly if not specified.\n\n')
    examples = [((f'{prog} /var/lib/scadere/certificates'
                  ' https://scadere.example/ :4433'),
                 ('serve renewal reminder feed using information'
                  ' from /var/lib/scadere/certificates on localhost:4433,'
                  ' to be reverse proxied to https://scadere.example/')),
                ('https://scadere.example/',
                 'feed for all checked TLS certificates'),
                ('https://scadere.example/?domain=example.com&domain=net',
                 ('feed for checked TLS certificates for example.com,'
                  ' its subdomains, and domains under the TLD NET'))]

    parser = ArgumentParser(prog=prog, allow_abbrev=False, description=desc,
                            epilog=format_epilog(examples),
                            formatter_class=GNUHelpFormatter)
    parser.add_argument('-v', '--version', action='version',
                        version=format_version('scadere-listen',
                                               '2025', 'Nguyễn Gia Phong'))
    parser.add_argument('-t', '--title', help='title of Atom feeds')
    parser.add_argument('certs', metavar='PATH', type=Path)
    parser.add_argument('base_url', metavar='URL', type=with_trailing_slash)
    parser.add_argument('netloc', metavar='[HOST][:PORT]', nargs='?',
                        type=NetLoc(None), default=('localhost', None))

    args = parser.parse_args(arguments)
    run(listen(args.certs, args.title,
               args.base_url, *args.netloc))  # pragma: no cover


if __name__ == '__main__':  # pragma: no cover
    main()



# Tests for the HTTP server
#
# SPDX-FileCopyrightText: 2025 Nguyễn Gia Phong
# SPDX-License-Identifier: AGPL-3.0-or-later

from asyncio import open_unix_connection, start_unix_server
from contextlib import asynccontextmanager, contextmanager
from copy import deepcopy
from datetime import datetime
from email.parser import BytesHeaderParser
from functools import partial
from http import HTTPMethod
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp, mkstemp
from urllib.parse import urljoin, urlsplit
from xml.etree.ElementTree import (XML, XMLParser, indent,
                                   tostring as str_from_xml)

from hypothesis import HealthCheck, given, settings
from hypothesis.strategies import (booleans, composite, data,
                                   datetimes, from_type, integers,
                                   sampled_from, sets, text, uuids)
from hypothesis.provisional import domains, urls
from pytest import raises

from scadere.check import base64_from_str, printable
from scadere.listen import (handle, is_subdomain, path, parse_summary,
                            str_from_base64, with_trailing_slash, xml)

ATOM_NAMESPACES = {'': 'http://www.w3.org/2005/Atom'}
XHTML_NAMESPACES = {'': 'http://www.w3.org/1999/xhtml'}


def ports():
    """Return a Hypothesis strategy for TCP ports."""
    return integers(1, 65535)


def serials():
    """Return a Hypothesis strategy for TLS serial number."""
    return integers(1, 256**20-1)


def base64s():
    """Return a Hypothesis strategy for printable strings in base64."""
    return text().filter(printable).map(base64_from_str)


@given(domains(), ports(), serials(), text())
def test_path(hostname, port, number, string):
    r = path(hostname, port, number, string).split('/')
    assert r[0] == hostname
    assert int(r[1]) == port
    assert int(r[3]) == number
    assert str_from_base64(r[2]) == string


@given(domains(), sets(domains()))
def test_is_subdomain(subject, objects):
    if not objects:
        assert is_subdomain(subject, objects)
    elif is_subdomain(subject, objects):
        assert any(child == '' or child.endswith('.')
                   for child in map(subject.removesuffix, objects))
    else:
        for obj in objects:
            assert (not subject.endswith(obj)
                    or not subject.removesuffix(obj).endswith('.'))


def xml_unsupported_type(child):
    """Check if child is of a type supported by the XML constructor."""
    return not isinstance(child, (tuple, str, int, datetime))


@given(text(), from_type(type).flatmap(from_type).filter(xml_unsupported_type))
def test_xml_unsupported_type(tag, child):
    with raises(AssertionError, match='Expected code to be unreachable'):
        xml((tag, {}, child))


@composite
def certificates(draw):
    """Return a Hypothesis strategy for certificate summaries."""
    valid = draw(booleans())
    not_before = draw(datetimes()).isoformat() if valid else 'N/A'
    not_after = draw(datetimes()).isoformat()
    hostname = draw(domains())
    port = draw(ports())
    number = draw(serials()) if valid else draw(uuids(version=4)).int
    string = draw(base64s())
    return f'{not_before} {not_after} {hostname} {port} {number} {string}'


@contextmanager
def tmp_cert_file(lines):
    cert_file = Path(mkstemp(text=True)[1])
    cert_file.write_text('\n'.join(lines))
    try:
        yield cert_file
    finally:
        cert_file.unlink()


def is_base_url(url):
    """Check if the given URL has a trailing slash.

    The parser for command-line arguments
    enforces this property for base URLs.
    """
    return url.endswith('/')


@given(urls())
def test_with_trailing_slash(url):
    if is_base_url(url):
        assert with_trailing_slash(url) == url
    else:
        assert with_trailing_slash(url) == f'{url}/'


def is_root(base_url, url):
    """Check if the given URL points to the same base URL.

    Paths starting with // are excluded because urljoin
    sometimes confuse them with URL scheme.
    """
    base_path = urlsplit(base_url).path
    url_path = urlsplit(url).path
    return urlsplit(urljoin(base_url, url_path)).path == base_path


def has_usual_path(url):
    """Check if the given URL path tends to mess with urljoin."""
    return is_root(url, url)


@asynccontextmanager
async def connect(socket):
    """Return a read-write stream for an asyncio TCP connection."""
    reader, writer = await open_unix_connection(socket)
    try:
        yield reader, writer
    finally:
        writer.close()
        await writer.wait_closed()


async def write_request(writer, request):
    """Write given request."""
    writer.write(request.encode())
    await writer.drain()
    assert writer.can_write_eof()
    writer.write_eof()


async def fetch_xml(socket, url, content_type):
    """Fetch the content at the URL from the socket."""
    header_parser = BytesHeaderParser()
    xml_parser = XMLParser(encoding='utf-8')
    async with connect(socket) as (reader, writer):
        await write_request(writer, f'GET {url} HTTP/1.1\r\n')
        status = await reader.readuntil(b'\r\n')
        assert status == b'HTTP/1.1 200 OK\r\n'
        headers_bytes = await reader.readuntil(b'\r\n\r\n')
        headers = header_parser.parsebytes(headers_bytes)
        assert headers['Content-Type'] == content_type
        content = await reader.read()
        assert len(content) == int(headers['Content-Length'])
        return XML(content.decode(), xml_parser)


def equal_xml(a, b):
    """Check if the two XML elements are equal."""
    a_copy, b_copy = deepcopy(a), deepcopy(b)
    indent(a_copy)
    indent(b_copy)
    return str_from_xml(a_copy).rstrip() == str_from_xml(b_copy).rstrip()


async def check_feed(socket, base_url):
    """Check the Atom feed at the given path and its entry pages."""
    feed = await fetch_xml(socket, base_url, 'application/atom+xml')
    for entry in feed.findall('entry', ATOM_NAMESPACES):
        link = entry.find('link', ATOM_NAMESPACES).attrib
        assert link['rel'] == 'alternate'
        page = await fetch_xml(socket, link['href'], link['type'])
        assert equal_xml(entry.find('.//dl', XHTML_NAMESPACES),
                         page.find('.//dl', XHTML_NAMESPACES))


async def check_server(handler, func, *args):
    """Test request handler using func."""
    d = Path(mkdtemp())
    socket = d / 'sock'
    try:
        async with await start_unix_server(handler, socket):
            await func(socket, *args)
    finally:
        rmtree(d)


def unique_netlocs(summaries):
    """Return summaries for unique network locations."""
    return {parse_summary(summary)[2:4]: summary
            for summary in summaries}.values()


@given(urls().filter(is_base_url).filter(has_usual_path),
       sets(certificates(), min_size=1).map(unique_netlocs),
       text().filter(printable))
@settings(suppress_health_check=[HealthCheck.too_slow])
async def test_content(base_url, certs, title):
    base_path = urlsplit(base_url).path
    with tmp_cert_file(certs) as cert_file:
        handler = partial(handle, cert_file, base_url, title=title)
        await check_server(handler, check_feed, base_path)


async def bad_request(socket, request):
    """Expect from socket a HTTP response with status 400."""
    async with connect(socket) as (reader, writer):
        await write_request(writer, request)
        status = await reader.readuntil(b'\r\n')
        assert status == b'HTTP/1.1 400 Bad Request\r\n'


@given(text().filter(lambda request: '\r\n' not in request))
async def test_incomplete_request(request):
    with tmp_cert_file(()) as cert_file:
        handler = partial(handle, cert_file, 'http://localhost')
        await check_server(handler, bad_request, request)


async def not_found(socket, url):
    """Send GET request for URL and expect HTTP status 404 from socket."""
    async with connect(socket) as (reader, writer):
        await write_request(writer, f'GET {url} HTTP/1\r\n')
        status = await reader.readuntil(b'\r\n')
        assert status == b'HTTP/1 404 Not Found\r\n'


@given(data())
@settings(suppress_health_check=[HealthCheck.too_slow])
async def test_unrecognized_url(drawer):
    base_url = drawer.draw(urls().filter(is_base_url), label='base URL')
    url = drawer.draw(urls().filter(lambda url: not is_root(base_url, url)),
                      label='request URL')
    with tmp_cert_file(()) as cert_file:
        handler = partial(handle, cert_file, base_url)
        await check_server(handler, not_found, urlsplit(url).path)


async def method_not_allowed(socket, method, url):
    """Expect from socket a HTTP response with status 405."""
    async with connect(socket) as (reader, writer):
        await write_request(writer, f'{method} {url} HTTP/1.1\r\n')
        status = await reader.readuntil(b'\r\n')
        assert status == b'HTTP/1.1 405 Method Not Allowed\r\n'


@given(urls(), sampled_from(HTTPMethod).filter(lambda method: method != 'GET'))
async def test_unallowed_method(base_url, method):
    with tmp_cert_file(()) as cert_file:
        handler = partial(handle, cert_file, base_url)
        await check_server(handler, method_not_allowed, method.value, base_url)


async def unsupported_http_version(socket, url, version):
    """Expect from socket a HTTP response with status 505."""
    async with connect(socket) as (reader, writer):
        await write_request(writer, f'GET {url} HTTP/{version}\r\n')
        status = await reader.readuntil(b'\r\n')
        assert status == b'HTTP/1.1 505 HTTP Version Not Supported\r\n'


@given(urls().filter(is_base_url).filter(has_usual_path), integers(10))
async def test_unsupported_http_version(base_url, version):
    with tmp_cert_file(()) as cert_file:
        handler = partial(handle, cert_file, base_url)
        await check_server(handler, unsupported_http_version,
                           base_url, version)



# Tests for the TLS client
#
# SPDX-FileCopyrightText: 2025 Nguyễn Gia Phong
# SPDX-License-Identifier: AGPL-3.0-or-later

from asyncio import get_running_loop, start_server
from datetime import datetime, timedelta, timezone
from io import StringIO
from ssl import Purpose, create_default_context as tls_context

from hypothesis import given
from pytest import mark
from trustme import CA

from scadere.check import base64_from_str, check, printable
from scadere.listen import parse_summary, str_from_base64

# Times in X.509 certificates are YYYYMMDDHHMMSSZ (RFC 5280)
SECONDS_AGO = datetime.now(timezone.utc).replace(microsecond=0)
NEXT_DAY = SECONDS_AGO + timedelta(days=1)
NEXT_WEEK = SECONDS_AGO + timedelta(days=7)


@given(...)
def test_base64(string: str):
    assert str_from_base64(base64_from_str(string)) == string


async def noop(reader, writer):
    """Do nothing."""
    writer.close()
    await writer.wait_closed()


async def get_cert_summary(netloc, after, ca):
    """Fetch TLS certificate expiration summary for netloc."""
    loop = get_running_loop()
    output = StringIO()
    await loop.run_in_executor(None, check, [netloc], after, output, ca)
    if not output.getvalue():
        return None
    summary, = map(parse_summary, output.getvalue().splitlines())
    return summary


@mark.parametrize('domain', ['localhost'])
@mark.parametrize('ca_name', ['trustme', '\x1f'])
@mark.parametrize('not_after', [SECONDS_AGO, NEXT_DAY, NEXT_WEEK])
@mark.parametrize('after', [NEXT_DAY, NEXT_WEEK])
@mark.parametrize('trust_ca', [False, True])
async def test_check(domain, ca_name, not_after, after, trust_ca):
    ctx = tls_context(Purpose.CLIENT_AUTH)
    ca = CA(organization_name=ca_name)
    cert = ca.issue_cert(domain, not_before=SECONDS_AGO, not_after=not_after)
    cert.configure_cert(ctx)
    ca.configure_trust(ctx)
    async with await start_server(noop, domain, ssl=ctx) as server:
        port = server.sockets[0].getsockname()[1]
        assert isinstance(port, int)
        summary = await get_cert_summary((domain, port), after,
                                         ca if trust_ca else None)
        if not trust_ca:
            assert summary[0] is None
            assert 'self-signed certificate' in summary[5]
        elif not_after == SECONDS_AGO:
            assert summary[0] is None
            assert 'certificate has expired' in summary[5]
        elif not printable(ca_name):
            assert summary[0] is None
            assert 'control character' in summary[5]
        elif not_after > after:
            assert summary is None
        else:
            assert summary[0] == SECONDS_AGO
            assert summary[1] == not_after
            assert summary[2] == domain
            assert summary[3] == port
            assert summary[5] == ca_name



# Tests for CLI help formatting
#
# SPDX-FileCopyrightText: 2025 Nguyễn Gia Phong
# SPDX-License-Identifier: AGPL-3.0-or-later

from contextlib import redirect_stdout, suppress
from io import StringIO

from hypothesis import example, given
from pytest import fixture, mark, raises

from scadere import EXAMPLE_PREFIX, EXAMPLE_DESCRIPTION_PREFIX, NetLoc
from scadere.check import main as check
from scadere.listen import main as listen


@fixture(scope='session')
def help_string(request):
    string = StringIO()
    with suppress(SystemExit), redirect_stdout(string):
        request.param(arguments=['--help'])
    return string.getvalue()


@mark.parametrize('help_string', [check, listen], indirect=True)
def test_usage_prefix(help_string):
    assert help_string.startswith('Usage: ')


@mark.parametrize('help_string', [check, listen], indirect=True)
def test_options_heading(help_string):
    assert '\n\nOptions:\n' in help_string


@mark.parametrize('help_string', [check], indirect=True)
def test_one_or_more(help_string):
    assert ' HOST[:PORT]...\n' in help_string


@mark.parametrize('help_string', [check], indirect=True)
@mark.parametrize('short,long,metavar', [('-d', '--days', 'DAYS'),
                                         ('-o', '--output', 'PATH')])
def test_long_option(help_string, short, long, metavar):
    assert f'{short} {metavar}, {long}={metavar}' in help_string


@mark.parametrize('help_string', [check, listen], indirect=True)
def test_examples(help_string):
    epilog_index = help_string.find('\n\nExamples:\n')
    assert epilog_index >= 0
    epilog = help_string[epilog_index:].removeprefix('\n\n')
    examples = epilog[:epilog.find('\n\n')]
    lines = examples.removeprefix('Examples:\n').splitlines()

    assert EXAMPLE_DESCRIPTION_PREFIX.startswith(EXAMPLE_PREFIX)
    assert not lines[0].startswith(EXAMPLE_DESCRIPTION_PREFIX)
    assert lines[-1].startswith(EXAMPLE_DESCRIPTION_PREFIX)

    must_be_desc = False
    for line in lines:
        if must_be_desc:
            assert line.startswith(EXAMPLE_DESCRIPTION_PREFIX)
            must_be_desc = False
        else:
            assert line.startswith(EXAMPLE_PREFIX)
            must_be_desc = not line.startswith(EXAMPLE_DESCRIPTION_PREFIX)


@example('a.example:b', None)  # string is unlikely to match .*:\D+
@example('a.example:98', None)  # string is unlikely to match .*:\d+
@given(...)
def test_netloc(string: str, default_port: int | None):
    netloc = NetLoc(default_port)
    if ':' not in string:
        assert netloc(string) == (string, default_port)
    else:
        hostname, port = string.rsplit(':', 1)
        try:
            port_number = int(port)
        except ValueError:
            with raises(ValueError):
                netloc(string)
        else:
            assert netloc(string) == (hostname, port_number)
